<!DOCTYPE html>
<html lang="en">

<head>
  <title><%= "#{title.capitalize} - " unless title == 'index' %>Sequel: The Database Toolkit for Ruby</title>
  <meta content="text/html; charset=utf-8" http-equiv="content-type" />
  
  <meta name="viewport" content="width=device-width, initial-scale=1, maximum-scale=1">
  <meta name="keywords" content="ruby, sequel, sql, toolkit, orm, plugins, documentation" />
  <meta name="description" content="Ruby Sequel is a lightweight database toolkit for Ruby." />

  <link rel="icon" type="image/icon" sizes="32x32" href="images/favicon.ico">
  <link href="css/application.css" media="screen" rel="stylesheet" type="text/css" />
  <link href="css/<%= title %>.css" media="screen" rel="stylesheet" type="text/css" />
</head>

<body class="body">
  <header class="header">
    <nav class="nav nav--header">
      <a href="index.html">
        <img src="images/ruby-sequel.svg" alt="Ruby Sequel Logo" class="nav__img nav__img--header" />
      </a>
      <ul class="nav__ul nav__ul--header">
        <li class="nav__li">
          <a href="documentation.html" class="nav__a">Documentation</a>
        </li>
        <li class="nav__li">
          <a href="plugins.html" class="nav__a">Plugins</a>
        </li>
        <li class="nav__li">
          <a href="development.html" class="nav__a">Development</a>
        </li>
        <li class="nav__li">
          <a href="links.html" class="nav__a">Links</a>
        </li>
      </ul>
      <form class="nav__form nav__form--header" action="https://www.google.com/search" autocomplete="off">
        <input type="hidden" name="sitesearch" value="sequel.jeremyevans.net" />
        <input class="nav__input" type="search" placeholder="Search docs, plugins, etc..." name="q" value="" />
      </form>
    </nav>
  </header>

  <main class="main">
    <div class="main__container">
      <%= content %>
    </div>
  </main>

  <footer class="footer">
    <nav class="nav">
      <a href="index.html">
        <img src="images/ruby-sequel-monocrome.svg" alt="Ruby Sequel Logo" class="nav__img nav__img--footer" />
      </a>
      <ul class="nav__ul">
        <li class="nav__li">
          <a href="documentation.html" class="nav__a">Documentation</a>
          <a href="documentation.html#general-overview" class="a">General Overview</a>
          <a href="documentation.html#databases" class="a">Databases</a>
          <a href="documentation.html#datasets" class="a">Datasets</a>
          <a href="documentation.html#models" class="a">Models</a>
          <a href="documentation.html#misc" class="a">Misc</a>
          <a href="documentation.html#rdoc" class="a">RDoc</a>
          <a href="documentation.html#release-notes" class="a">Release Notes</a>
        </li>
        <li class="nav__li">
          <a href="plugins.html" class="nav__a">Plugins</a>
          <a href="plugins.html#model-plugins" class="a">Model Plugins</a>
          <a href="plugins.html#sequel-extensions" class="a">Sequel Extensions</a>
        </li>
        <li class="nav__li">
          <a href="development.html" class="nav__a">Development</a>
          <a href="development.html#bugs" class="a">Reporting Bugs</a>
          <a href="development.html#contribute" class="a">Contribute</a>
          <a href="development.html#code" class="a">Source Code</a>
          <a href="development.html#license" class="a">License</a>
        </li>
        <li class="nav__li">
          <a href="links.html" class="nav__a">Links</a>
          <a href="links.html#articles" class="a">Articles</a>
          <a href="links.html#presentations" class="a">Presentations</a>
          <a href="links.html#logos" class="a">Logos &amp; Badges</a>
          <a href="links.html#using-sequel" class="a">Using Sequel</a>
          <a href="https://rubygems.org/gems/sequel/" class="a">RubyGems</a>
          <a href="https://github.com/jeremyevans/sequel/" class="a">GitHub</a>
        </li>
      </ul>
      <form action="https://www.google.com/search" autocomplete="off">
        <input type="hidden" name="sitesearch" value="sequel.jeremyevans.net" />
        <input class="nav__input" type="search" placeholder="Search docs, plugins, etc..." name="q" value="" />
      </form>
    </nav>
  </footer>
</body>
</html>
