<section class="development">

<h1 class="h1">Development</h1>

<hr class="hr" />

<h2 class="h2">Sequel is being actively developed. <span class="h2__span">New versions of Sequel are generally released monthly on the first of the month. You can join in on the discussions, ask questions, suggest features, and discuss Sequel in general by asking questions in <a class="a" href="https://github.com/jeremyevans/sequel/discussions">GitHub Discussions</a>. There are many years of past discussions on the searchable <a class="a" href="https://www.mail-archive.com/sequel-talk@googlegroups.com/">sequel-talk mailing list archive</a>.</span></h2>

<hr class="hr" />

<div class="row row--grid">
<div class="row__item">
<a name="bugs"></a>

<h3 class="h3">Reporting Bugs</h3>
<p class="p">To report a bug in Sequel, use <a class="a" href="https://github.com/jeremyevans/sequel/issues">GitHub Issues</a>.  If you aren't sure if something is a bug, post a question on <a class="a" href="https://github.com/jeremyevans/sequel/discussions">GitHub Discussions</a>.</p>
<p class="p">Note that GitHub Issues should not be used to ask questions about how to use Sequel, use <a class="a" href="https://github.com/jeremyevans/sequel/discussions">GitHub Discussions</a>.</p>

</div>

<div class="row__item">
<a name="contributing"></a>

<h3 class="h3">Contributing</h3>
<p class="p">The easiest way to contribute is to use git, post the changes to a public repository, and send a pull request, either via GitHub or the Google Group.  Posting patches to the bug tracker or the Google Group works fine as well.</p>
<p class="p">You can find the <a class="a" href="https://github.com/jeremyevans/sequel/blob/master/CONTRIBUTING">contributing guideliness</a> on GitHub.</p>

</div>

<div class="row__item">
<a name="code"></a>

<h3 class="h3">Source Code</h3>
<p class="p">The master source code repository is <a class="a" href="https://github.com/jeremyevans/sequel/">jeremyevans/sequel</a> on GitHub.</p>

</div>

<div class="row__item">
<a name="license"></a>

<h3 class="h3">License</h3>
<p class="p">Sequel is distributed under the <a class="a" href="rdoc/files/MIT-LICENSE.html">MIT License</a>.  Patches are assumed to be submitted under the same license as Sequel.</p>

</div>
</div>
</section>
