<section class="links">
  <h1 class="h1">Links</h1>

  <hr class="hr" />

  <a name="articles"></a>
  <h2 class="h2">Articles</h2>

  <div class="row">
    <div class="row__item">
      <h3 class="h3">Web</h3>
      <ul class="ul">
        <li class="ul__li">
          <a class="a" href="http://rubydev.ru/2011/10/interview_jeremy_evans_about_sequel_ru/">rubydev.ru: Interview with Jeremy Evans about Sequel and not only</a>
          October 2011, Russian Translation
        </li>
        <li class="ul__li">
          <a  class="a" href="http://on-ruby.blogspot.com/2009/01/sequel-interview-with-jeremy-evans.html">On Ruby: Sequel Interview with Jeremy Evans</a>
          January 2009
        </li>
        <li class="ul__li">
          <a  class="a" href="http://www.infoq.com/news/2008/11/sequel-ruby-db-toolkit">InfoQ: Sequel Interview with Jeremy Evans</a>
          November 2008
        </li>
        <li class="ul__li">
          <a  class="a" href="http://on-ruby.blogspot.com/2008/01/sequel-interview-with-sharon-rosner.html">On Ruby: Sequel Interview with Sharon Rosner</a>
          January 2008
        </li>
        <li class="ul__li">
          <a  class="a" href="http://www.oreillynet.com/ruby/blog/2007/07/ruby_project_spotlight_june_07.html">O"Reilly Ruby: Ruby Project Spotlight: Sequel</a>
          July 2007
        </li>
      </ul>
    </div>
    <div class="row__item">
      <h3 class="h3">Print</h3>
      <ul class="ul">
        <li class="ul__li">
          <a  class="a" href="https://web.archive.org/web/20110209051845/http://www.impressjapan.jp/books/2579">Ruby! Vol. 1</a>
          In japanese (thanks Shunichi Arai)
        </li>
      </ul>
    </div>
  </div>

  <hr class="hr" />

  <a name="presentations"></a>
  <h2 class="h2">Presentations</h2>
  <div class="row">
    <div class="row__item">
      <ul class="ul">
        <li class="ul__li"><strong>RubyDay 2020</strong> <a class="a" href="https://speakerdeck.com/janko_m/sequel-when-activerecord-is-not-enough">Sequel: When Active Record is not enough</a></li>
        <li class="ul__li"><strong>RubyC 2014</strong> <a class="a" href="http://code.jeremyevans.net/presentations/rubyc2014-2/index.html">Give-and-Go with PostgreSQL and Sequel</a> <a class="a parens" href="http://www.youtube.com/watch?v=toAcnwqlU1Q">Video</a></li>
        <li class="ul__li"><strong>RubyC 2014</strong> <a class="a" href="http://code.jeremyevans.net/presentations/rubyc2014-1/index.html">Deep Dive Into Eager Loading Limited Associations</a> <a class="a parens" href="http://www.youtube.com/watch?v=KbF7RMk_2Qo">Video</a></li>
        <li class="ul__li"><strong>RailsClub 2013</strong> <a class="a" href="http://code.jeremyevans.net/presentations/railsclub2013/index.html?trans=no">Give-and-Go with PostgreSQL and Sequel</a> <a class="a parens" href="http://code.jeremyevans.net/presentations/railsclub2013/index.html">In Russian</a></li>
        <li class="ul__li"><strong>Heroku May 2012</strong> <a class="a" href="http://code.jeremyevans.net/presentations/heroku201205/index.html">The Development of Sequel</a></li>
        <li class="ul__li"><strong>PostgreSQL Conference West 2011</strong> <a class="a" href="http://code.jeremyevans.net/presentations/pgwest2011/index.html">Sequel: The Database Toolkit for Ruby</a></li>
        <li class="ul__li"><strong>Lone Star Ruby Conference 2009</strong> <a class="a" href="http://code.jeremyevans.net/presentations/lsrc2009_presentation/sequel-lsrc2009.html">Sequel: SQL in Ruby</a></li>
        <li class="ul__li"><strong>RubyKaigi 2009</strong> <a class="a" href="http://code.jeremyevans.net/presentations/rk2009_presentation/sequel-rubykaigi2009.html">Sequel: SQL in Ruby</a></li>
        <li class="ul__li"><strong>LA Ruby Conf 2009</strong> <a class="a" href="http://code.jeremyevans.net/presentations/larc2009_presentation/sequel-larc2009-pres.html">Sequel: The Database Toolkit for Ruby</a> <a class="a parens" href="https://www.youtube.com/watch?v=0Euba97UEdI">Video</a></li>
        <li class="ul__li"><strong>MountainWest RubyConf 2009</strong> <a class="a parens" href="http://code.jeremyevans.net/presentations/mwrc2009_presentation.html">Sequel: The Database Toolkit for Ruby</a> <a class="a" href="https://www.youtube.com/watch?v=yCjtdfv9Z_0">Video</a> <a class="a parens" href="http://code.jeremyevans.net/presentations/mwrc2009_presentation.txt">Transcript</a></li>
      </ul>
    </div>
  </div>

  <hr class="hr" />

  <a name="logos"></a>
  <h2 class="h2">Logos &amp; Badges</h2>
  <ul class="assets">
    <li class="assets__item">
      <a class="assets__container" href="images/ruby-sequel.svg">
        <img src="images/ruby-sequel.svg" alt="Sequel's logo in color" class="assets__img" />
      </a>
      <p class="assets__description">Sequel logo SVG (Color)</p>
    </li>
    <li class="assets__item">
      <a class="assets__container" href="images/ruby-sequel.svg">
        <img src="images/ruby-sequel-monocrome.svg" alt="Sequel's logo in black" class="assets__img" />
      </a>
      <p class="assets__description">Sequel logo SVG (Monocrome)</p>
    </li>
    <li class="assets__item">
      <a class="assets__container" href="images/sequel-button.png">
        <img src="images/sequel-button.png" alt="Sequel's repo badge" class="assets__img" />
      </a>
      <p class="assets__description">Repository Badge</p>
    </li>
  </ul>

  <hr class="hr" />

  <h2 class="h2">Additional Links</h2>
  <div class="row row--grid">
    <div class="row__item">
      <a name="github"></a>
      <h3 class="h3">GitHub</h3>
      <ul class="ul">
        <li class="ul__li">
          <a href="https://github.com/jeremyevans/sequel/" class="a">Sequel on GitHub</a>
        </li>
      </ul>
    </div>
    <div class="row__item">
      <a name="rubygems"></a>
      <h3 class="h3">RubyGems</h3>
      <ul class="ul">
        <li class="ul__li">
          <a href="https://rubygems.org/gems/sequel/" class="a">Sequel on RubyGems</a>
        </li>
      </ul>
    </div>
    <div class="row__item">
      <a name="blog"></a>
      <h3 class="h3">Blog</h3>
            <ul class="ul">
        <li class="ul__li">
          <a href="https://sequel.jeremyevans.net/blog.html" class="a">The Sequel Blog (no longer updated)</a>
        </li>
      </ul>
    </div>
  </div>
</section>
