<section class="documentation">
<h1 class="h1">Documentation for Sequel (v<%= Sequel.version %>)</h1>

<hr class="hr" />

<h2 class="h2">General Info, Guides, Examples, and Tutorials</h2>

<div class="row row--grid">
  <div class="row__item">
    <a name="general-overview"></a>

    <h3 class="h3">General Overviews</h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/files/README_rdoc.html">README</a> Introduction to Sequel, start here if new</li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/security_rdoc.html">Security</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/cheat_sheet_rdoc.html">Cheat Sheet</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/sql_rdoc.html">Sequel for SQL Users</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/object_model_rdoc.html">Sequel Object Model</a></li>
    </ul>
  </div>
  <div class="row__item">
    <a name="databases"></a>

    <h3 class="h3">Databases</h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/files/doc/opening_databases_rdoc.html">Connecting to a Database</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/transactions_rdoc.html">Database Transactions</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/schema_modification_rdoc.html">Schema Modification</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/migration_rdoc.html">Migrations</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/sharding_rdoc.html">Primary/Replica Database Configurations and Sharding</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/postgresql_rdoc.html">PostgreSQL Specific Support</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/mssql_stored_procedures_rdoc.html">Microsoft SQL Server Stored Procedure Support</a></li>
    </ul>
  </div>
  <div class="row__item">
    <a name="datasets"></a>

    <h3 class="h3">Datasets</h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/files/doc/dataset_basics_rdoc.html">Dataset Basics</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/dataset_filtering_rdoc.html">Dataset Filtering</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/querying_rdoc.html">Querying in Sequel</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/prepared_statements_rdoc.html">Prepared Statements/Bound Variables</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/virtual_rows_rdoc.html">Virtual Row Blocks</a></li>
    </ul>
  </div>
  <div class="row__item">
    <a name="models"></a>

    <h3 class="h3">Models</h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/files/doc/mass_assignment_rdoc.html">Model Mass Assignment</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/validations_rdoc.html">Model Validations</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/model_hooks_rdoc.html">Model Hooks</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/model_dataset_method_design_rdoc.html">Model Dataset Method Design</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/association_basics_rdoc.html">Model Associations Basics and Options</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/advanced_associations_rdoc.html">Advanced Model Associations</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/model_plugins_rdoc.html">Model Plugins</a></li>
    </ul>
  </div>
  <div class="row__item">
    <a name="misc"></a>

    <h3 class="h3">Misc.</h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/files/doc/code_order_rdoc.html">Code Order</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/reflection_rdoc.html">Reflection</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/thread_safety_rdoc.html">Thread Safety</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/fork_safety_rdoc.html">Fork Safety</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/testing_rdoc.html">Testing With Sequel</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/extensions_rdoc.html">Sequel Extensions</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/core_extensions_rdoc.html">Core Extensions</a></li>
      <li class="ul__li"><a class="a" href="rdoc/files/doc/bin_sequel_rdoc.html">bin/sequel</a></li>
    </ul>
  </div>
</div>

<hr class="hr" />

<a name="rdoc"></a>
<h2 class="h2">RDoc</h2>

<div class="row row--grid">
  <div class="row__item">

    <h3 class="h3"><a class="a" href="rdoc/index.html">Core / Model</a></h3>

    <ul class="ul">
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel.html">Sequel</a> Main module</li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Database.html">Database</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Dataset.html">Dataset</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Model.html">Model</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Model/ClassMethods.html">Model Class Methods</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Model/InstanceMethods.html">Model Instance Methods</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Model/Associations/ClassMethods.html">Model Association Class Methods</a></li>
      <li class="ul__li"><a class="a" href="rdoc/classes/Sequel/Model/Associations/DatasetMethods.html">Model Eager Loading Methods</a></li>
    </ul>
  </div>
  <div class="row__item">
    <h3 class="h3"><a class="a" href="rdoc-adapters/index.html">Adapters</a></h3>
  </div>
  <div class="row__item">
    <h3 class="h3"><a class="a" href="rdoc-plugins/index.html">Extensions/Plugins</a></h3>
  </div>
</div>

<hr class="hr" />

<div class="row">
  <a name="release-notes"></a>
  <h3 class="h3"><a class="a" href="rdoc/files/CHANGELOG.html">Release Notes</a></h3>
  <ul class="ul">
    <% %w'5 4 3'.each do |i| %>
      <li><strong>Sequel <%= i %></strong>
        <ul class="release-notes">
          <% 
            lines = []
            Dir["../doc/release_notes/#{i}.*.txt"].map{|f| File.basename(f)}.each do |f|
              (lines[f.split('.')[1].to_i/10] ||= []) << f
            end
            lines.reverse.each do |fs|
          %>
            <li>
              <% fs.sort_by{|f| f.split('.').map{|x| x.to_i}}.reverse.each do |f| %>
                <a class="a" href="rdoc/files/doc/release_notes/<%= f.gsub('.', '_') %>.html"><%= f.sub(/\.txt$/, '').sub(/(..)\.0$/, '\\1') %></a> |
              <% end %>
            </li>
          <% end %>
        </ul>
      </li>
    <% end %>
    <% %w'2 1'.each do |i| %>
      <li><strong>Sequel <%= i %></strong>
        <ul class="release-notes">
          <li>
            <% Dir["../doc/release_notes/#{i}.*.txt"].map{|f| File.basename(f)}.sort_by{|f| f.split('.').map{|x| x.to_i}}.reverse.each do |f| %>
              <a class="a" href="rdoc/files/doc/release_notes/<%= f.gsub('.', '_') %>.html"><%= f.sub(/\.txt$/, '').sub(/(..)\.0$/, '\\1') %></a> |
            <% end %>
          </li>
        </ul>
      </li>
    <% end %>
  </ul>
</div>

<hr class="hr" />

<div class="row row--grid">
  <div class="row__item">
    <h3 class="h3"><a class="a" href="rdoc/files/CHANGELOG.html">Change Log</a></h3>
  </div>
  <div class="row__item">
    <h3 class="h3"><a class="a" href="rdoc/files/MIT-LICENSE.html">License</a></h3>
  </div>
</div>
</section>
